/*
	EASY-ROB Development V2.3

	Stefan Anton

	Copyright (c) 1996,1997, 1998

	modul: er_dvlp.h
*/
#ifndef ER_DVLP_INCLUDE
#define ER_DVLP_INCLUDE

#define DllExport __declspec( dllexport )

/*******************************
*   User Vector for Testing    *
*******************************/
extern float *inq_sys_data_v_usr0(void);
extern float *inq_sys_data_v_usr1(void);
/*******************************
*   kinematics & dynammics     *
*******************************/

typedef struct tframe {
	float  n[3],o[3],a[3],p[3];
} FRAME,frame;

#define MAXSTR 128
#define PI 3.14159265359
#define DEG (180.0/PI)
#define RAD (PI/180.0)

#define KIN_CONFIGS    8 /* number of  configurations */
#define KIN_DOFS       12 /* number DOF's */
#define KIN_PASSIV_JNTS 12 /* number of passiv Joints */

#define KIN_EPSI 0.5e-6

#define WARN_OK 0
#define WARN_SINGULAR 1
#define WARN_UNREACH 2
#define WARN_CNFG 3
#define WARN_NO_INVKIN 4
#define WARN_SWE_EXCEED 5

#define ROT_X 1
#define ROT_Y 2
#define ROT_Z 3

typedef struct {
	float theta,d,a,alfa;
} ROB_DH;

void  vec_to_frame (float *q,frame *T);
void  frame_to_vec (float *q,frame *T);


void  T_mal_T (FRAME *To, FRAME *Ti1, FRAME *Ti2 );
void  R_mal_R (FRAME *To, FRAME *Ti1, FRAME *Ti2 );
void  inv_T (FRAME *To, FRAME *Ti);
void  mul_invT_T (FRAME *To, FRAME *Ti1, FRAME *Ti2);
void  mul_T_invT (FRAME *To, FRAME *Ti1, FRAME *Ti2);
void  incr_T (FRAME *, FRAME *, int);
void  incr_R (FRAME *, FRAME *, int);
void  cpy_R (FRAME *, FRAME *);
void mul_T_pos (float *,FRAME *,float *);

void   rob_kin_T_mal_T ( frame *Tin1,frame *Tin2);
void  rob_kin_R_mal_R ( frame *Tin1,frame *Tin2);
void  rob_kin_frame_ident(frame *T);
void   rob_kin_rot (int rot_idx,double q, frame *T);
void  rob_kin_trans (frame *T,double x, double y, double z);
void   dh_to_frame(ROB_DH *rdh,frame *T);
int  rob_kin_joint_trans(int joint_dof);
int  rob_kin_joint_rot(int joint_dof);
float  rob_kin_to_DEG(int joint_dof);
float  rob_kin_to_RAD(int joint_dof);
float  sssa(float s1,float s2,float s3);
float  sasssa(float s1,float beta,float s2,float s3,float s4);

frame *  inq_T(void);
frame *  inq_bT0(void);
frame *  inq_tTw(void);
float *  inq_q_solut(void);
float *  inq_v_solut(void);
frame *  inq_bTw(void);

double  rob_kin_atan2(double,double);
void  rob_kin_set_warnings(int);
void  rob_kin_user_msg(char *s);

double *  inq_achs_length(void);
double *  inq_achs_offsets1(void);
double *  inq_achs_offsets2(void);
double *  inq_achs_rotx(void);
double *  inq_achs_roty(void);
double *  inq_achs_rotz(void);
frame *  inq_achs_T(void);
double *  inq_sing_tol(void);
double *  inq_joint_offset(void);
float *  inq_q(int soln);
int *  inq_warnings(void);
int *  inq_verpolt(void);
int  inq_num_dofs(void);
int  inq_num_passiv_jnts(void);

int  inq_num_configs(void);
int  inq_config(void);

ROB_DH *  inq_rob_dh_activ(void);
ROB_DH *  inq_rob_dh_passiv(void);
float *  inq_homepos(void);
float *  inq_swe_max(void);
float * inq_swe_min(void);
float * inq_v_max(void);
float * inq_a_max(void);
float * inq_t_max(void);

char *  inq_robotname(void);
char *  inq_kin_direction_activ(void);
char *  inq_kin_type_activ(void);

float *  inq_bdxw(void);
float *  inq_dq(void);

char *  inq_kin_direction_passiv(void);
char *  inq_kin_type_passiv(void);
char *  inq_kin_chain_type(void);
char *  inq_kin_calc(int passiv_jnt_no);
int *  inq_kin_attach_dof(void);
frame *  inq_kin_achs_T_passiv(int passiv_jnt_no);
frame *  inq_kin_T_passiv(int passiv_jnt_no);

int moni_msg( char *fmt, ... );
void moni_T (char *s,frame *T);
void moni_v (char *s,float *v,int n,float scal);
void moni_i (char *s,int *v,int n);
void moni_s (char *s,short *v,int n);
void ErrorExitBox(char *s);
void MsgBox(char *s);

double *inq_univ_trans_tol_rad(void);
double *inq_univ_trans_tol_m(void);
int *inq_univ_trans_ilimit(void);
double *inq_univ_trans_mask(void);
double *inq_univ_trans_weight(void);
frame **inq_univ_trans_T(void);

/********************
*  dynamics         *
********************/

float  inq_dt_cntrl(void);
float  inq_dt_sys(void);

float * inq_cntrl_kpp(void);
float * inq_cntrl_kpv(void);
float * inq_cntrl_kiv(void);
float * inq_cntrl_uc_kiv1(void);
float * inq_cntrl_aux1(void);
float * inq_cntrl_aux2(void);
float * inq_cntrl_aux3(void);
float * inq_cntrl_e1(void);
float * inq_cntrl_e2(void);

float * inq_model_N(void);
float * inq_model_T_base(void);
float * inq_model_theta_base(void);
float * inq_model_d_base(void);
float * inq_model_vz1_v(void);
float * inq_model_vz1_a(void);
float * inq_model_z1(void);
float * inq_model_z2(void);
float *  inq_uc(void);
float *  inq_uc1(void);
float *  inq_q_dyn(void);
float *  inq_v_dyn(void);
frame *  inq_bTw_dyn(void);


/********************
*  motion planner   *
********************/

// Return codes for Interpolation
#define IPO_OK 0
#define IPO_IN_POSITION 1
#define IPO_ERROR -1

typedef enum {JOINT, CP, CIRC, IPO_AUX} IPO_MODE;
typedef enum {VARIABLE, FIX, TANGENTIAL, AUX, VARIABLE2, QUATERNION} CIRC_ORI_MODE;
typedef enum {LIN_VARIABLE, LIN_FIX, LIN_TANGENTIAL, LIN_AUX, LIN_QUATERNION} LIN_ORI_MODE;

float *inq_ipo_exec_qs(void);
float *inq_ipo_exec_vs(void);
frame *inq_ipo_exec_Ts(void);

float *inq_ipo_path_vq(void);
float *inq_ipo_path_vqs(void);
float *inq_ipo_path_vqe(void);
float *inq_ipo_path_vx(void);
float *inq_ipo_path_vxs(void);
float *inq_ipo_path_vxe(void);
float *inq_ipo_path_aq(void);
float *inq_ipo_path_ax(void);
float *inq_ipo_path_ov_pro(void);
CIRC_ORI_MODE *inq_ipo_path_circ_ori_mode(void);
LIN_ORI_MODE *inq_ipo_path_lin_ori_mode(void);

int *inq_ipo_prep_n_dof(void);
float *inq_ipo_prep_t_bahn(void);
float *inq_ipo_prep_strt_q(void);
float *inq_ipo_prep_end_q(void);
float *inq_ipo_prep_dq(void);
frame *inq_ipo_prep_strt_T(void);
frame *inq_ipo_prep_end_T(void);
frame *inq_ipo_prep_inter_T(void);
frame *inq_ipo_prep_center_T(void);
float *inq_ipo_prep_dx(void);

IPO_MODE *inq_ipo_data_ipo_mode(void);

float *inq_ipo_data_gtime(void);
float *inq_ipo_data_ctime(void);
float *inq_ipo_data_dt_ipo(void);
float *inq_ipo_data_lead_time(void);
float *inq_ipo_data_lag_time(void);
int *inq_ipo_data_ipo_in_position(void);
float *inq_ipo_data_old_q(void);
float *inq_ipo_data_new_q(void);
frame *inq_ipo_data_old_T(void);
frame *inq_ipo_data_new_T(void);

int circ_center_point(float *p1,float *p2,float *p3,frame *pTc,float *radius,float *phi);

#endif
