/*
	EASY-ROB Development V2.3

	3 * 2 * 12 user defined motion planner routines (Joint, CP, Circular)
	for Path Preparation and Path Execution are supported.

	int ipo_xxx_prep_user_1..12 (void)
	void ipo_xxx_exec_user_1..12 (float time)

	xxx is "jnt" | "cp" | "circ"

	The EASY-ROB base program "easyrobx.exe" provides more than 100 functions.

	Stefan Anton

	Copyright (c) 1996,1997, 1998

	modul: ipo_user.c

	To build a new er_ipo.dll, just execute the do_dlls.bat batch file in
	directory .\er_dvlp\.
*/
#include "stdafx.h"

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <malloc.h>

#include "..\er_dvlp\er_dvlp.h"

#define MIN_IPO_TIME 0.001

#pragma warning(disable : 4244)     // float conversation

DllExport int er_ipo(void)
{
	return(0);
}
/*
	User defined Structur (example) for Interpolator User_xx
*/
#define IPO_DATA_FILE "ipo.dat"
typedef struct {
	float a;
	FILE *fp;
	char buff[MAXSTR];
} IPO_USR_xx;

static int ipo_ini_xx(void **ipo_usr_ptr)
/* ipo_ini_xx()
	allocates memory for the user defined struture IPO_USR_XX;
	return 1 -> error, no memory available
			 0 -> OK
*/
{
	IPO_USR_xx *ipo_ptr;
	if (*ipo_usr_ptr==NULL) {
		moni_msg("ipo_ini_xx() *ipo_usr_ptr=NULL -> malloc");
		*ipo_usr_ptr = (void *)malloc(sizeof(IPO_USR_xx));
		ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;
		if (ipo_ptr==NULL) {
			moni_msg("Error, malloc IPO_USR_xx in ipo_ini_xx()");
			return(1);
		}
		/*intialize values*/
		ipo_ptr->a=1.0;
	}
	return(0);
}
static int ipo_chk_xx(void **ipo_usr_ptr)
/* ipo_chk_xx()
	will check if the pointer *ipo_usr_ptr is NULL or in other words,
	if memory was allocated before
	return 1 -> error
			 0 -> OK
*/
{
	if (*ipo_usr_ptr==NULL) {
		moni_msg("Error in ipo_chk_xx() ipo_usr_ptr==NULL");
		return(1);
	}
	return(0);
}

static void my_out_vector_proc(FILE *fp,float time,float *v,int n)
/* my_out_vector_proc()
	is a local static user routine and will write n float values 'v'
	to the open stream 'fp'.
*/
{
	int i;
	fprintf(fp,"%8.3f",time);
	for (i=0;i<n;i++) fprintf(fp,"%9.4f",v[i]);
	fprintf(fp,"\n");
	return;
}

static int ipo_prep_template(void **ipo_usr_ptr)
{
	*inq_ipo_prep_t_bahn()=0;
	if (*ipo_usr_ptr==NULL) return(1);
	return(0);
}
static int ipo_exec_template(float t,void **ipo_usr_ptr)
{
	if (*ipo_usr_ptr==NULL) return(0);
	if (t) return(0);
	return(0);
}
/***************************************************************
			 Joint Motion Planner
***************************************************************/
DllExport int  ipo_jnt_prep_user_1(void **ipo_usr_ptr)
/* example: geometric joint motion planner, preparation
	return 1 -> robot is in target joint values, or
					an error occured
			 0 -> joint path is planned,the
					ipo_jnt_exec_user_x() is called

	Explanation:

	This function has to prepare the PTP (Point to Point) move
	from the joint start values "inq_ipo_prep_strt_q()" [rad] or [m]to
	the joint end values "inq_ipo_prep_end_q()" [rad] or [m].
	In this case we have a simple zero order velocity profil.

	Input values are:

	- The number of Degrees of Freedom of this robot "*inq_ipo_prep_n_dof()".
	- The nominal allowed joint speed "*inq_ipo_path_vq()" in [deg].
	- The programmed override "*inq_ipo_path_ov_pro()" in [%].

	Output values are (used in function ipo_jnt_exec_user_1):

	- t_bahn "*inq_ipo_prep_t_bahn()" is the required time in [sec]
	  for the PTP move.
	- The Difference Vector dq - "inq_ipo_prep_dq()" in [rad].
*/
{
	int i,n_dof;
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	float vq,t,*dq,*end_q,*strt_q,*t_bahn;

	// alloc memory and define your user defined structure
	if (ipo_ini_xx(ipo_usr_ptr)) return(1);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;
	if (ipo_ptr->a==0) {;} //do nothing with it

	/* open a file stream to WRITE values */
	if ((ipo_ptr->fp=fopen(IPO_DATA_FILE,"w"))==NULL) {
		return(1); /* error */
	}
	fprintf(ipo_ptr->fp,"ipo_jnt_prep_user_1\n");

	// Output to file moni.msg
	rob_kin_user_msg("ipo_jnt_prep_user_1()");

	// get pointers and/or values from the easyrob data base
	dq = inq_ipo_prep_dq();
	end_q = inq_ipo_prep_end_q();
	strt_q = inq_ipo_prep_strt_q();
	n_dof = *inq_ipo_prep_n_dof();
	t_bahn = inq_ipo_prep_t_bahn();

	vq= *inq_ipo_path_vq()*RAD * *inq_ipo_path_ov_pro()*0.01;
	*t_bahn=0.0;

	// calculate the joint difference dq = end_q - start_q
	// calculate the need time (temporary) t = dq / vq
	// save the largest time in *t_bahn
	// "bahn is german and means path"
	for(i=0;i<n_dof;i++) {
		dq[i] = end_q[i] - strt_q[i];
		t = fabs(dq[i])/vq;
		if (t> *t_bahn) *t_bahn=t;
	}
	// If the time "*t_bahn" is less then MIN_IPO_TIME (see header er_dvlp.h)
	// the robot is in the target joint values "start_q"
	// Else reduce dq by the leading joint axis
	if (*t_bahn>MIN_IPO_TIME) {
		for(i=0;i<n_dof;i++) dq[i] /= *t_bahn;
		return(0);
	}
	return(1); // robot is in the target joint values "start_q"
} /*End of ipo_jnt_prep_user_1*/

DllExport int  ipo_jnt_exec_user_1(float t,void **ipo_usr_ptr)
/* example: geometric joint motion planner, execution
	Input: t is the current time in [sec]
	return 0 -> new values are calculated or kept constant.
					ipo_jnt_exec_user_1() is called until
					the current time t > *t_bahn (see preparation)
	return 1 -> new values are calculated or kept constant
					ipo_jnt_exec_user_1() is called independant
					from the relation of the current time t to *t_bahn.
					Thus it is called until return is set to 0.

	Explanation:
*/
{
	int i,n_dof;
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	float *q,*v,*dq,*end_q,*strt_q,*t_bahn;

	if (ipo_chk_xx(ipo_usr_ptr)) return(0);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;

//	moni_msg("ipo_jnt_exec_user_1() time %7.4f",t);

	v = inq_ipo_exec_vs();
	q = inq_ipo_exec_qs();

	t_bahn = inq_ipo_prep_t_bahn();
	dq = inq_ipo_prep_dq();
	end_q = inq_ipo_prep_end_q();
	strt_q = inq_ipo_prep_strt_q();
	n_dof = inq_num_dofs();

	if (t<0)
		for(i=0;i<n_dof;i++) { q[i] = strt_q[i]; v[i]=0; }
	else if (t<= *t_bahn) {
		for(i=0;i<n_dof;i++) {
			v[i] = dq[i];
			q[i] = strt_q[i]+t*dq[i];
		}
		my_out_vector_proc(ipo_ptr->fp,t,q,n_dof);
	}
	else {
		for(i=0;i<n_dof;i++) { q[i] = end_q[i]; v[i]=0; }
		my_out_vector_proc(ipo_ptr->fp,t,q,n_dof);
		moni_msg("close fp in ipo_exec_user_1");
		fclose(ipo_ptr->fp);
	}

	return(0);
} /*End of ipo_jnt_exec_user_1*/

DllExport int  ipo_jnt_prep_user_2(void **ipo_usr_ptr)
/* example: joint motion planner, preparation
	return 1 -> robot is in target joint values, or
					an error occured
			 0 -> joint path is planned, the
					ipo_jnt_exec_user_x() is called

	Explanation:
*/
{
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	float *t_bahn=inq_ipo_prep_t_bahn();

	// alloc memory and define your user defined structure
	if (ipo_ini_xx(ipo_usr_ptr)) return(1);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;
	if (ipo_ptr->a==0) {;} //do nothing with it

	/* open a file stream to read values */
	if ((ipo_ptr->fp=fopen(IPO_DATA_FILE,"r"))==NULL) {
		return(1); /* error */
	}
	fgets(ipo_ptr->buff,MAXSTR,ipo_ptr->fp);

	rob_kin_user_msg("ipo_jnt_prep_user_2()");
	*t_bahn=0;
	return(0);
} /*End of ipo_jnt_prep_user_1*/

DllExport int  ipo_jnt_exec_user_2(float t,void **ipo_usr_ptr)
/* example: joint motion planner, execution
	Input: t is the current time in [sec]

	return 0 -> new values are calculated or kept constant.
					ipo_jnt_exec_user_1() is called until
					the current time t > *t_bahn (see preparation)
	return 1 -> new values are calculated or kept constant
					ipo_jnt_exec_user_1() is called independant
					from the relation of the current time t to *t_bahn.
					Thus it is called until return is set to 0.
*/
{
	int i,n_dof,ret=0;
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	float *q,*v,*strt_q;

	if (ipo_chk_xx(ipo_usr_ptr)) return(0);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;

	//moni_msg("ipo_jnt_exec_user_2() time %7.4f",t);

	v = inq_ipo_exec_vs();
	q = inq_ipo_exec_qs();
	strt_q = inq_ipo_prep_strt_q();
	n_dof = inq_num_dofs();

	if (t<0 || t>2.0) {
		for(i=0;i<n_dof;i++) { q[i] = strt_q[i]; v[i]=0; }
		fclose(ipo_ptr->fp);
		moni_msg("time %8.3f close fp",t);
	}
	else {
		fgets(ipo_ptr->buff,MAXSTR,ipo_ptr->fp);
		moni_msg("time %-8.3f %s",t,ipo_ptr->buff);

		for(i=0;i<n_dof;i++) { v[i] = 0.0;
			q[i] = strt_q[i]+t*0.5;
		}
		ret=1; /* call ipo_jnt_exec again unitl
					 eof for example or t>2 in this case */
	}
	return(ret);
} /*End of ipo_jnt_exec_user_2*/
DllExport int  ipo_jnt_prep_user_3(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_3(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_4(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_4(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_5(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_5(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_6(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_6(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_7(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_7(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_8(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_8(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_9(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_9(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_10(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_10(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_11(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_11(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_jnt_prep_user_12(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_jnt_exec_user_12(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}

/***************************************************************
			 CP Motion Planner
***************************************************************/
DllExport int ipo_cp_prep_user_1(void **ipo_usr_ptr)
/* example: geometric cp motion planner, preparation
	return 1 -> robot is in target location
			 0 -> cp path is planned
*/
{
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	int i;
	float vq,vx,t,vec[6];
	float *t_bahn,*dx;
	frame *dT,T,*end_T,*strt_T;

	moni_msg("ipo_cp_prep_user_1()");

	// alloc memory and define your user defined structure
	if (ipo_ini_xx(ipo_usr_ptr)) return(1);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;
	if (ipo_ptr->a==0) {;} //do nothing with it

	dT = &T;

	t_bahn = inq_ipo_prep_t_bahn();
	strt_T = inq_ipo_prep_strt_T();
	end_T = inq_ipo_prep_end_T();
	dx = inq_ipo_prep_dx();

	vq= *inq_ipo_path_vq()*RAD * *inq_ipo_path_ov_pro()*0.01;
	vx= *inq_ipo_path_vx() * *inq_ipo_path_ov_pro()*0.01;

	/* position */
	*t_bahn=0;

	for(i=0;i<3;i++) {
		dx[i] = end_T->p[i] - strt_T->p[i];
		t = fabs(dx[i])/vx;
		if (t> *t_bahn) *t_bahn=t;
	}

	/* orientierung */
	mul_invT_T(dT,strt_T,end_T);
	frame_to_vec(vec,dT);

	for(i=3;i<6;i++) {
		t = fabs((dx[i]=vec[i]))/vq;
		if (t> *t_bahn) *t_bahn=t;
	}

	/* dx reduzieren */
	if (*t_bahn>MIN_IPO_TIME) {
		for(i=0;i<6;i++) dx[i] /= *t_bahn;
		return(0);
	}
	return(1);
}
DllExport int ipo_cp_exec_user_1(float t,void **ipo_usr_ptr)
/* example: geometric joint motion planner, execution
	Input: t is the current time in [sec]
*/
{
	int i;
	float vec[6],*v;
	float *t_bahn,*dx;

	frame *dT,*bTw,T,*end_T,*strt_T;

	if (ipo_chk_xx(ipo_usr_ptr)) return(0);

	moni_msg("ipo_cp_exec_user_1() time %7.4f",t);

	dT = &T;

	bTw = inq_ipo_exec_Ts();
	v = inq_ipo_exec_vs();

	t_bahn = inq_ipo_prep_t_bahn();
	strt_T = inq_ipo_prep_strt_T();
	end_T = inq_ipo_prep_end_T();
	dx = inq_ipo_prep_dx();

	for(i=0;i<inq_num_dofs();i++) v[i] = 0.0;

	for(i=0;i<3;i++) vec[i] = 0;
	for(i=3;i<6;i++) vec[i] = t*dx[i];
	vec_to_frame(vec,dT);
	R_mal_R(bTw,strt_T,dT);
	if (t<0) *bTw = *strt_T;
	else if (t<= *t_bahn)
		for(i=0;i<3;i++) bTw->p[i] = strt_T->p[i]+t*dx[i];
	else *bTw = *end_T;

	return(0);
}
DllExport int  ipo_cp_prep_user_2(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_2(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_3(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_3(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_4(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_4(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_5(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_5(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_6(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_6(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_7(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_7(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_8(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_8(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_9(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_9(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_10(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_10(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_11(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_11(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_cp_prep_user_12(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_cp_exec_user_12(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}

/***************************************************************
			 CIRC Motion Planner
***************************************************************/

DllExport int ipo_circ_prep_user_1(void **ipo_usr_ptr)
/* example: geometric cp motion planner, preparation
	return 1 -> robot is in target location
			 0 -> cp path is planned
*/
{
	IPO_USR_xx *ipo_ptr; /* pointer to your user defined structure */
	int i;
	float vq,vx,t,vec[6];
	float *t_bahn,*dx;
	frame *dT,T,*end_T,*strt_T;

	moni_msg("ipo_circ_prep_user_1()");

	// alloc memory and define your user defined structure
	if (ipo_ini_xx(ipo_usr_ptr)) return(1);
	ipo_ptr = (IPO_USR_xx *)*ipo_usr_ptr;
	if (ipo_ptr->a==0) {;} //do nothing with it

	dT = &T;

	t_bahn = inq_ipo_prep_t_bahn();
	strt_T = inq_ipo_prep_strt_T();
	end_T = inq_ipo_prep_end_T();
	dx = inq_ipo_prep_dx();

	vq= *inq_ipo_path_vq()*RAD * *inq_ipo_path_ov_pro()*0.01;
	vx= *inq_ipo_path_vx() * *inq_ipo_path_ov_pro()*0.01;

	/* position */
	*t_bahn=0;

	for(i=0;i<3;i++) {
		dx[i] = end_T->p[i] - strt_T->p[i];
		t = fabs(dx[i])/vx;
		if (t> *t_bahn) *t_bahn=t;
	}

	/* orientierung */
	mul_invT_T(dT,strt_T,end_T);
	frame_to_vec(vec,dT);

	for(i=3;i<6;i++) {
		t = fabs((dx[i]=vec[i]))/vq;
		if (t> *t_bahn) *t_bahn=t;
	}

	/* dx reduzieren */
	if (*t_bahn>MIN_IPO_TIME) {
		for(i=0;i<6;i++) dx[i] /= *t_bahn;
		return(0);
	}
	return(1);
}
DllExport int ipo_circ_exec_user_1(float t,void **ipo_usr_ptr)
/* example: geometric joint motion planner, execution
	Input: t is the current time in [sec]
*/
{
	int i;
	float vec[6],*v;
	float *t_bahn,*dx;

	frame *dT,*bTw,T,*end_T,*strt_T;

	if (ipo_chk_xx(ipo_usr_ptr)) return(0);

	moni_msg("ipo_circ_exec_user_1() time %7.4f",t);

	dT = &T;

	bTw = inq_ipo_exec_Ts();
	v = inq_ipo_exec_vs();

	t_bahn = inq_ipo_prep_t_bahn();
	strt_T = inq_ipo_prep_strt_T();
	end_T = inq_ipo_prep_end_T();
	dx = inq_ipo_prep_dx();

	for(i=0;i<inq_num_dofs();i++) v[i] = 0.0;

	for(i=0;i<3;i++) vec[i] = 0;
	for(i=3;i<6;i++) vec[i] = t*dx[i];
	vec_to_frame(vec,dT);
	R_mal_R(bTw,strt_T,dT);
	if (t<0) *bTw = *strt_T;
	else if (t<= *t_bahn)
		for(i=0;i<3;i++) bTw->p[i] = strt_T->p[i]+t*dx[i];
	else *bTw = *end_T;

	return(0);
}

typedef struct { // user defined structure
	float rs,hs,phis; // start radius, -height, -angle 
	float re,he,phie; // end radius, -height, -angle 
	frame dT; // helper frame needed for ipo execution
	CIRC_ORI_MODE circ_ori_mode; // ... needed for ipo execution
} IPO_USR_CIRC;

static int ipo_ini_circ(void **ipo_usr_ptr)
/* ipo_ini_circ()
	allocates memory for the user defined struture IPO_USR_CIRC;
	return 1 -> error, no memory available
			 0 -> OK
*/
{
	IPO_USR_CIRC *ipo_ptr;
	if (*ipo_usr_ptr==NULL) {
		moni_msg("ipo_ini_circ() *ipo_usr_ptr=NULL -> malloc");
		*ipo_usr_ptr = (void *)malloc(sizeof(IPO_USR_CIRC));
		ipo_ptr = (IPO_USR_CIRC *)*ipo_usr_ptr;
		if (ipo_ptr==NULL) {
			moni_msg("Error, malloc IPO_USR_xx in ipo_ini_circ()");
			return(1);
		}
		/*intialize values*/
		ipo_ptr->rs=ipo_ptr->re=0.0;
		ipo_ptr->hs=ipo_ptr->he=0.0;
		ipo_ptr->phis=ipo_ptr->phie=0.0;
		rob_kin_frame_ident(&(ipo_ptr->dT));
		ipo_ptr->circ_ori_mode=VARIABLE;
		moni_msg("ipo_ini_circ() done");
	}
	return(0);
}
DllExport int  ipo_circ_prep_user_2(void **ipo_usr_ptr)
{
	IPO_USR_CIRC *ipo_ptr; /* pointer to your user defined structure */

	int i,
		is_circle; // 1 - if circle else 0
	float vq,vx,t,vec[6];
	frame *dT,T,*dRphi,R;
	float *p1,*p2,*p3;
	float radius,phi;
	frame *pTc,*end_T,*strt_T;
	float *t_bahn,*dx;

	moni_msg("ipo_circ_prep_user_2()");

	// alloc memory and define your user defined structure
	if (ipo_ini_circ(ipo_usr_ptr)) return(1);
	ipo_ptr = (IPO_USR_CIRC *)*ipo_usr_ptr;

	t_bahn = inq_ipo_prep_t_bahn();
	strt_T = inq_ipo_prep_strt_T();
	end_T = inq_ipo_prep_end_T();
	dx = inq_ipo_prep_dx();

	dT = &T;
	dRphi = &R;

	// joint- and cartesian speeds
	vq= *inq_ipo_path_vq()*RAD * *inq_ipo_path_ov_pro()*0.01;
	vx= *inq_ipo_path_vx() * *inq_ipo_path_ov_pro()*0.01;

	/* position */
	*t_bahn=0;
	is_circle = 0; // idx if circle

	pTc = inq_ipo_prep_center_T(); // center of circle
	p1 = inq_ipo_prep_strt_T()->p; // start point
	p2 = inq_ipo_prep_inter_T()->p; // via point
	p3 = inq_ipo_prep_end_T()->p; // end point of circle

	// calculate the center of circle 
	// radius and angle of circle
	if (circ_center_point(p1,p2,p3,pTc,&radius,&phi)) {
		is_circle = 0; // p1-p3 do not define a circle
		*t_bahn=0;
	} else {
		is_circle = 1; // p1-p3 define a circle
		ipo_ptr->rs = ipo_ptr->re = radius;
		ipo_ptr->phie = phi;

		dx[0] = (ipo_ptr->re - ipo_ptr->rs);
		dx[1] = (ipo_ptr->phie - ipo_ptr->phis);
		dx[2] = (ipo_ptr->he - ipo_ptr->hs); // currently not used

		// t_bahn
		*t_bahn = fabs(0.5*(ipo_ptr->re+ipo_ptr->rs)*dx[1])/vx;
	}

	/* calculate the orientation */
	ipo_ptr->circ_ori_mode = *inq_ipo_path_circ_ori_mode(); // cpy ori mode
	for(i=3;i<6;i++) vec[i]=0.0;

	switch (ipo_ptr->circ_ori_mode) {
	case TANGENTIAL:
		// calculate the end orientation
		// ipo_ptr->dT is from center of circle to start point p1
		mul_invT_T(&(ipo_ptr->dT),pTc,strt_T);
		rob_kin_rot(ROT_Z,(double)phi,dRphi);
		incr_R(pTc,dRphi,2);
		R_mal_R(end_T,dRphi,&(ipo_ptr->dT)); // modified target orient.
		break;
	case VARIABLE: // standard simple orientation interpolation
		mul_invT_T(dT,strt_T,end_T);
		frame_to_vec(vec,dT);
		break;
	case FIX:
		cpy_R(end_T,strt_T); // modified target orient.
	case AUX:
	case VARIABLE2: 
	case QUATERNION:
	default:
		break;
	}

	for(i=3;i<6;i++) {
		t = fabs((dx[i]=vec[i]))/vq;
		if (t>*t_bahn) {
			*t_bahn=t;
		}
	}

	/* dx reduction */
	if (*t_bahn>MIN_IPO_TIME) {
		for(i=0;i<3;i++) dx[i] /= *t_bahn; // r,phi,h
		for(i=3;i<6;i++) dx[i] /= *t_bahn; // A B C
		return(IPO_OK); // OK --> ipo_exec()
	}

	// check circle idx
	if (is_circle==0) {
		return(IPO_ERROR); // no circle
	}

	return(IPO_IN_POSITION); // in position
}
DllExport int  ipo_circ_exec_user_2(float t,void **ipo_usr_ptr)
/*
	will execute the prepared circle
	t is the current time of the path
*/
{
	IPO_USR_CIRC *ipo_ptr;  /* pointer to your user defined structure */
	int i;
	float *v,phi,r,h,vec[6];
	frame *dT,T,*dRphi,R;
	frame *bTw,*pTc,*end_T,*strt_T;
	float *t_bahn,*dx;

	if (!ipo_chk_xx(ipo_usr_ptr)) {
		// get pointer of user defined structure
		ipo_ptr = (IPO_USR_CIRC *)*ipo_usr_ptr;
	} else return(0);

	dT = &T;
	dRphi = &R;
	// get more pointers
	bTw = inq_ipo_exec_Ts(); // ptr. to desired robot lacation
	pTc = inq_ipo_prep_center_T(); // ptr to center of prepared circle

	t_bahn = inq_ipo_prep_t_bahn(); // ptr to path time
	strt_T = inq_ipo_prep_strt_T(); // ptr to start location
	end_T = inq_ipo_prep_end_T(); // ptr to end location
	dx = inq_ipo_prep_dx();

	v = inq_ipo_exec_vs(); /*reset nom joint speed */
	for(i=0;i<inq_num_dofs();i++) v[i] = 0.0;
	
	// calc. angle, raduis and height in dependency of time and dx
	r   = dx[0] * t + ipo_ptr->rs;
	phi = dx[1] * t + ipo_ptr->phis;
	h   = dx[2] * t + ipo_ptr->hs;

	switch (ipo_ptr->circ_ori_mode) {
	case TANGENTIAL:
		if (t<=*t_bahn) {
		    rob_kin_rot(ROT_Z,(double)phi,dRphi);
		    incr_R(pTc,dRphi,2);
		    R_mal_R(bTw,dRphi,&(ipo_ptr->dT));
		}
		break;
	case VARIABLE:
	case VARIABLE2:
		// simple orientation
		for(i=0;i<3;i++) vec[i] = 0;
		for(i=3;i<6;i++) vec[i] = t*dx[i];
		vec_to_frame(vec,dT);
		R_mal_R(bTw,strt_T,dT);
		break;
	case FIX:
	case AUX:
	case QUATERNION:
	default:
		cpy_R(bTw,strt_T);
		break;
	}

	// position of circ ipo
	if (t<0) *bTw = *strt_T;
	else if (t<=*t_bahn) {
		vec[0]= r * cos(phi);
		vec[1]= r * sin(phi);
		vec[2]= h;
		mul_T_pos(bTw->p,pTc,vec);
	}
	else *bTw = *end_T;
	
	return(0);
}
DllExport int  ipo_circ_prep_user_3(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_3(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_4(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_4(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_5(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_5(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_6(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_6(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_7(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_7(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_8(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_8(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_9(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_9(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_10(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_10(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_11(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_11(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
DllExport int  ipo_circ_prep_user_12(void **ipo_usr_ptr)
{
	return(ipo_prep_template(ipo_usr_ptr));
}
DllExport int  ipo_circ_exec_user_12(float t,void **ipo_usr_ptr)
{
	return(ipo_exec_template(t,ipo_usr_ptr));
}
